"""The SHV RPC type Bool."""

from __future__ import annotations

import typing

from .. import SHVType
from ..value import is_shvbool
from .base import RpcType


class RpcTypeBool(RpcType):
    """The Bool type representation."""

    __obj = None

    def __new__(cls) -> RpcTypeBool:  # noqa: D102
        if cls.__obj is None:
            cls.__obj = object.__new__(cls)
        return cls.__obj

    def __str__(self) -> str:
        return "b"

    @staticmethod
    def validate(value: SHVType) -> typing.TypeGuard[bool]:  # noqa: D102
        return is_shvbool(value)


rpctype_bool = RpcTypeBool()
"""The singleton for :class:`RpcTypeBool`."""



"""Abstraction of the SHV RPC device alert."""

import datetime
import enum
import typing

from .value import SHVIMapType, SHVType


class RpcAlert:
    """Device alert representation."""

    NOTICE_MIN: int = 0
    """The minimal value for the Notice level."""
    NOTICE_MAX: int = 20
    """The maximal value for the Notice level."""
    WARNING_MIN: int = 21
    """The minimal value for the Warning level."""
    WARNING_MAX: int = 42
    """The maximal value for the Warning level."""
    ERROR_MIN: int = 43
    """The minimal value for the Error level."""
    ERROR_MAX: int = 63
    """The maximal value for the Error level."""

    class Key(enum.IntEnum):
        """Key inm the alert IMap."""

        DATE = 0
        LEVEL = 1
        ID = 2
        INFO = 3

    def __init__(self, value: SHVIMapType | None = None) -> None:
        self.value: dict[int, SHVType] = dict(value) if value else {}

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RpcAlert) and self.id == other.id

    @property
    def date(self) -> datetime.datetime:
        """Date and time of the alert creation."""
        value = self.value.get(self.Key.DATE)
        if not isinstance(value, datetime.datetime):
            raise ValueError(f"Must be datetime but is: {type(value)}")
        return value

    @date.setter
    def date(self, value: datetime.datetime) -> None:
        self.value[self.Key.DATE] = value

    @property
    def level(self) -> int:
        """Alert level."""
        value = self.value.get(self.Key.LEVEL)
        if not isinstance(value, int):
            raise ValueError(f"Must be int but is: {type(value)}")
        return value

    @level.setter
    def level(self, value: int) -> None:
        if value < self.NOTICE_MIN or value > self.ERROR_MAX:
            raise ValueError("Must be between 0 and 63")
        self.value[self.Key.LEVEL] = value

    @property
    def id(self) -> str:
        """Alert identifier."""
        value = self.value.get(self.Key.ID)
        if not isinstance(value, str):
            raise ValueError(f"Must be string but is: {type(value)}")
        return value

    @id.setter
    def id(self, value: str) -> None:
        self.value[self.Key.ID] = value

    @classmethod
    def new(
        cls, level: int, alert_id: str, date: datetime.datetime | None = None
    ) -> typing.Self:
        """Create a new alert."""
        return cls({
            cls.Key.DATE: date or datetime.datetime.now(),
            cls.Key.LEVEL: level,
            cls.Key.ID: alert_id,
        })



"""Python implementation of Silicon Heaven."""

from .__version__ import VERSION
from .chainpack import ChainPack, ChainPackReader, ChainPackWriter
from .cpon import Cpon, CponReader, CponWriter
from .path import SHVPath, SHVPathParents
from .rpcaccess import RpcAccess
from .rpcalert import RpcAlert
from .rpcdir import RpcDir
from .rpcerrors import (
    RpcError,
    RpcErrorCode,
    RpcInvalidParamError,
    RpcLoginRequiredError,
    RpcMethodCallExceptionError,
    RpcMethodNotFoundError,
    RpcNotImplementedError,
    RpcRequestInvalidError,
    RpcTryAgainLaterError,
    RpcUserIDRequiredError,
)
from .rpcfile import (
    FileProvider,
    FileProviderAppend,
    FileProviderFixedSize,
    FileProviderRO,
    FileProviderRW,
    RpcFile,
    RpcFileStat,
)
from .rpclogin import RpcLogin, RpcLoginType
from .rpcmessage import RpcMessage
from .rpcparam import SHVGetKey, shvarg, shvargt, shvget, shvgett, shvt
from .rpcri import rpcri_legacy_subscription, rpcri_match, rpcri_relative_to
from .rpctransport import (
    RpcClient,
    RpcClientPipe,
    RpcClientTCP,
    RpcClientTTY,
    RpcClientUnix,
    RpcClientWebSockets,
    RpcProtocolBlock,
    RpcProtocolSerial,
    RpcProtocolSerialCRC,
    RpcServer,
    RpcServerTCP,
    RpcServerTTY,
    RpcServerUnix,
    RpcServerWebSockets,
    RpcServerWebSocketsUnix,
    RpcTransportProtocol,
    connect_rpc_client,
    create_rpc_server,
    init_rpc_client,
)
from .rpcurl import RpcProtocol, RpcUrl
from .shvbase import SHVBase
from .shvclient import SHVClient
from .shvdevice import SHVDevice
from .shvmethods import SHVMethods
from .shvvalueclient import SHVValueClient
from .shvversion import SHV_VERSION, SHV_VERSION_MAJOR, SHV_VERSION_MINOR
from .value import (
    SHVBool,
    SHVBoolType,
    SHVBytes,
    SHVDatetime,
    SHVDecimal,
    SHVFloat,
    SHVIMap,
    SHVIMapType,
    SHVInt,
    SHVList,
    SHVListType,
    SHVMap,
    SHVMapType,
    SHVMeta,
    SHVMetaType,
    SHVNull,
    SHVNullType,
    SHVStr,
    SHVType,
    SHVUInt,
    decimal_rexp,
    is_shvbool,
    is_shvimap,
    is_shvlist,
    is_shvmap,
    is_shvnull,
    is_shvtype,
    shvmeta,
    shvmeta_eq,
)

__all__ = [
    "SHV_VERSION",
    "SHV_VERSION_MAJOR",
    "SHV_VERSION_MINOR",
    "VERSION",
    "ChainPack",
    "ChainPackReader",
    "ChainPackWriter",
    "Cpon",
    "CponReader",
    "CponWriter",
    "FileProvider",
    "FileProviderAppend",
    "FileProviderFixedSize",
    "FileProviderRO",
    "FileProviderRW",
    "RpcAccess",
    "RpcAlert",
    "RpcClient",
    "RpcClientPipe",
    "RpcClientTCP",
    "RpcClientTTY",
    "RpcClientUnix",
    "RpcClientWebSockets",
    "RpcDir",
    "RpcError",
    "RpcErrorCode",
    "RpcFile",
    "RpcFileStat",
    "RpcInvalidParamError",
    "RpcLogin",
    "RpcLoginRequiredError",
    "RpcLoginType",
    "RpcLoginType",
    "RpcMessage",
    "RpcMethodCallExceptionError",
    "RpcMethodNotFoundError",
    "RpcNotImplementedError",
    "RpcProtocol",
    "RpcProtocolBlock",
    "RpcProtocolSerial",
    "RpcProtocolSerialCRC",
    "RpcRequestInvalidError",
    "RpcServer",
    "RpcServerTCP",
    "RpcServerTTY",
    "RpcServerUnix",
    "RpcServerWebSockets",
    "RpcServerWebSocketsUnix",
    "RpcTransportProtocol",
    "RpcTryAgainLaterError",
    "RpcUrl",
    "RpcUserIDRequiredError",
    "SHVBase",
    "SHVBool",
    "SHVBoolType",
    "SHVBytes",
    "SHVClient",
    "SHVDatetime",
    "SHVDecimal",
    "SHVDevice",
    "SHVFloat",
    "SHVGetKey",
    "SHVIMap",
    "SHVIMapType",
    "SHVInt",
    "SHVList",
    "SHVListType",
    "SHVMap",
    "SHVMapType",
    "SHVMeta",
    "SHVMetaType",
    "SHVMethods",
    "SHVNull",
    "SHVNullType",
    "SHVPath",
    "SHVPathParents",
    "SHVStr",
    "SHVType",
    "SHVUInt",
    "SHVValueClient",
    "connect_rpc_client",
    "create_rpc_server",
    "decimal_rexp",
    "init_rpc_client",
    "is_shvbool",
    "is_shvimap",
    "is_shvlist",
    "is_shvmap",
    "is_shvnull",
    "is_shvtype",
    "rpcri_legacy_subscription",
    "rpcri_match",
    "rpcri_relative_to",
    "shvarg",
    "shvargt",
    "shvget",
    "shvgett",
    "shvmeta",
    "shvmeta_eq",
    "shvt",
]



"""The SHV RPC type oneof."""

from __future__ import annotations

import collections.abc
import typing

from .. import SHVType
from .base import RpcType
from .null import rpctype_null


class RpcTypeOneOf(RpcType):
    """The type that provides validation based on the multiple other types.

    If at least one of the provided types validates value then this type is
    valid.
    """

    def __init__(self, *args: RpcType) -> None:
        self._types = tuple(self.__inititer(args))

    @classmethod
    def __inititer(
        cls, types: collections.abc.Iterable[RpcType]
    ) -> collections.abc.Iterator[RpcType]:
        for tp in types:
            if isinstance(tp, RpcTypeOneOf):
                yield from cls.__inititer(tp.types)
            else:
                yield tp

    @property
    def types(self) -> tuple[RpcType, ...]:
        """Types this 'one of' allows."""
        return self._types

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RpcTypeOneOf) and self._types == other._types

    def __str__(self) -> str:
        return "|".join(str(t) for t in self._types)

    def validate(self, value: SHVType) -> typing.TypeGuard[float]:  # noqa: D102
        return any(t.validate(value) for t in self._types)


class RpcTypeOptional(RpcTypeOneOf):
    """The variant of :py:class:`RpcTypeOneOf` that implicitly adds :data:`rpctype_null`."""

    def __init__(self, *args: RpcType) -> None:
        if rpctype_null not in args:
            super().__init__(*args, rpctype_null)
        else:
            super().__init__(*args)



"""The base for the other RPC Types."""

from __future__ import annotations

import abc

from .. import SHVType


class RpcType(abc.ABC):
    """The base for the RPC Type specifications."""

    @abc.abstractmethod
    def validate(self, value: SHVType) -> bool:
        """Check if given value is of this type."""



"""The SHV RPC type bitfield."""

from __future__ import annotations

import collections.abc
import typing

from .. import SHVType
from .base import RpcType
from .bool import RpcTypeBool
from .enum import RpcTypeEnum
from .unsigned import RpcTypeUnsigned


class RpcTypeBitfield(RpcType, collections.abc.Sequence[tuple[int, RpcType, str]]):
    """The Bitfield type representation."""

    def __init__(self, *items: tuple[int, RpcType, str]) -> None:
        self._items = sorted(items, key=lambda v: v[0])
        self._mask = 0
        end = -1
        for i in self._items:
            if i[0] < 0:
                raise ValueError("Bit index can't be negative number")
            if i[0] <= end:
                raise ValueError(f"Bit {i[0]} is used in multiple items")
            size = self.bitsize(i[1])
            if size is None:
                raise ValueError(f"Type '{i[1]}' can't be in Bitfield")
            end = i[0] + size - 1
            self._mask |= 2**size - 1 << i[0]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RpcTypeBitfield) and self._items == other._items

    def __str__(self) -> str:
        defs = (
            f"{tp}:{key}" if i is None else f"{tp}:{key}:{i}"
            for tp, key, i in self.__seqiter(self._items)
        )
        return f"u[{','.join(defs)}]"

    @classmethod
    def __seqiter(
        cls,
        iiter: collections.abc.Iterable[tuple[int, RpcType, str]],
    ) -> collections.abc.Iterator[tuple[RpcType, str, int | None]]:
        expected = 0
        for i, tp, key in iiter:
            yield tp, key, None if expected == i else i
            size = cls.bitsize(tp)
            assert isinstance(size, int)
            expected = i + size

    @typing.overload
    def __getitem__(self, i: int, /) -> tuple[int, RpcType, str]: ...

    @typing.overload
    def __getitem__(
        self, i: slice, /
    ) -> collections.abc.Sequence[tuple[int, RpcType, str]]: ...

    def __getitem__(
        self, i: int | slice
    ) -> tuple[int, RpcType, str] | collections.abc.Sequence[tuple[int, RpcType, str]]:
        return self._items[i]

    def __len__(self) -> int:
        return len(self._items)

    def validate(self, value: SHVType) -> typing.TypeGuard[int]:  # noqa: D102
        return (
            isinstance(value, int)
            and value ^ value & self._mask == 0
            and all(
                True
                if isinstance(tp, RpcTypeBool)
                else tp.validate(self.extract(value, s, tp))
                for s, tp, _ in self
            )
        )

    @staticmethod
    def bitsize(tp: RpcType) -> int | None:
        """Calculate bit size required for this type if supported.

        :param tp: The type bit size should be deduced for.
        :return: Number bits to be used in the bitfield or ``None`` if it can't
          be included.
        """
        match tp:
            case RpcTypeBool():
                return 1
            case RpcTypeUnsigned():
                if tp.maximum is not None and tp.maximum != tp.minimum:
                    return (tp.maximum - tp.minimum).bit_length()
            case RpcTypeEnum():
                return max(tp).bit_length()
        return None

    @classmethod
    def extract(cls, value: int, start: int, tp: RpcType) -> int:
        """Extract integer based on the provided type and start.

        The handling is based on the type:
        * ``b``: ``1`` is returned for true and ``0`` for false.
        * ``u(MAX)``: number is returned as extracter.
        * ``u(MIN,MAX)``: number is returned with added ``MIN``.
        * ``i[]``: number is returned as extracted.
        """
        size = cls.bitsize(tp)
        if size is None:
            raise ValueError(f"Type '{tp} can't be in Bitfield")
        result: int = value >> start & (2**size - 1)
        if isinstance(tp, RpcTypeUnsigned) and tp.minimum > 0:
            result += tp.minimum
        return result

    @classmethod
    def deposit(cls, value: int, start: int, tp: RpcType, update: int = 0) -> int:
        """Insert integer based on the provided type and start.

        The handling is based on the type:
        * ``b``: odd numbers are inserted as true and even numbers as false.
        * ``u(MAX)``: number is inserted as is.
        * ``u(MIN,MAX)``: number is inserted minus the ``MIN``.
        * ``i[]``: number is inserted as is.
        """
        if isinstance(value, bool):
            value = 1 if value else 0
        elif not tp.validate(value):
            raise ValueError(f"Value '{value}' is invalid for type '{tp}'")
        elif isinstance(tp, RpcTypeUnsigned) and tp.minimum > 0:
            value -= tp.minimum
        size = cls.bitsize(tp)
        if size is None:
            raise ValueError(f"Type '{tp} can't be in Bitfield")
        mask: int = 2**size - 1
        return (value & mask) << start | update ^ mask << start



"""Check implementation of SimpleDevice."""

import datetime
import time

import pytest

from shv import (
    SHV_VERSION_MAJOR,
    SHV_VERSION_MINOR,
    VERSION,
    RpcAccess,
    RpcAlert,
    RpcDir,
    RpcMethodNotFoundError,
    RpcNotImplementedError,
    SHVDevice,
    shvmeta,
)


class Device(SHVDevice):
    DEVICE_NAME = "testdev"
    DEVICE_VERSION = "0.0.x"


@pytest.fixture(name="device")
async def fixture_device(shvbroker, url_test_device):
    """Run device and provide instance to access it."""
    device = await Device.connect(url_test_device)
    yield device
    await device.disconnect()


@pytest.mark.parametrize(
    "path,method,result",
    (
        ("test/device/.app", "shvVersionMajor", SHV_VERSION_MAJOR),
        ("test/device/.app", "shvVersionMinor", SHV_VERSION_MINOR),
        ("test/device/.app", "name", "pyshv-device"),
        ("test/device/.app", "version", VERSION),
        ("test/device/.app", "ping", None),
        ("test/device/.device", "name", "testdev"),
        ("test/device/.device", "version", "0.0.x"),
        ("test/device/.device", "serialNumber", None),
    ),
)
async def test_call(client, device, path, method, result):
    """Check that we can call various methods using blocking call."""
    res = await client.call(path, method)
    assert res == result
    assert shvmeta(res) == shvmeta(result)


async def test_call_date(client, device):
    """Check that we can call various methods using blocking call."""
    res = await client.call("test/device/.app", "date")
    assert isinstance(res, datetime.datetime)
    assert res.tzinfo is not None
    diff = datetime.datetime.now().astimezone() - res
    assert datetime.timedelta() <= diff < datetime.timedelta(seconds=1)


async def test_invalid_call(client):
    with pytest.raises(RpcMethodNotFoundError):
        await client.call(".app", "someInvalid")


@pytest.mark.parametrize(
    "path,result",
    (
        ("test/device", [".app", ".device"]),
        ("test/device/.app", []),
        ("test/device/.device", ["alerts"]),
        ("test/device/.device/alerts", []),
    ),
)
async def test_ls(client, device, path, result):
    """Verify that we can use ls method."""
    res = await client.ls(path)
    assert res == result


@pytest.mark.parametrize(
    "path,result",
    (
        (
            "test/device/.device",
            [
                RpcDir.stddir(),
                RpcDir.stdls(),
                RpcDir.getter("name", "n", "s", access=RpcAccess.BROWSE),
                RpcDir.getter("version", "n", "s", access=RpcAccess.BROWSE),
                RpcDir.getter("serialNumber", "n", "s|n", access=RpcAccess.BROWSE),
                RpcDir.getter("uptime", "n", "u|n", access=RpcAccess.BROWSE),
                RpcDir("reset", access=RpcAccess.COMMAND),
            ],
        ),
        (
            "test/device/.device/alerts",
            [
                RpcDir.stddir(),
                RpcDir.stdls(),
                RpcDir.getter(result="[!alert]", signal=True),
            ],
        ),
    ),
)
async def test_dir(client, device, path, result):
    """Verify that we can use dir method."""
    res = await client.dir(path)
    assert res == result


async def test_reset(client, device):
    """Check if reset can be called and we get the not implemented error."""
    with pytest.raises(RpcNotImplementedError):
        await client.call("test/device/.device", "reset")


async def test_uptime(client, device):
    """We can't test uptime in general but at least be reasonable with it."""
    uptime = int(time.monotonic())
    assert await client.call("test/device/.device", "uptime") >= uptime


async def test_alerts(value_client, device):
    """Check that we correctly manage alerts."""
    await value_client.subscribe("test/device/.device/alerts:get:chng")
    assert await value_client.prop_get("test/device/.device/alerts") == []

    now1 = datetime.datetime.now().replace(microsecond=0, tzinfo=datetime.UTC)
    test1 = RpcAlert.new(RpcAlert.WARNING_MIN, "Test1", now1)
    change = value_client.wait_for_change("test/device/.device/alerts")
    await device.change_alerts()  # Just to ensure that we do not send signal
    await device.change_alerts(add=test1)
    alerts = await change
    assert len(alerts) == 1
    assert alerts[0] == test1.value

    now2 = datetime.datetime.now().replace(microsecond=0, tzinfo=datetime.UTC)
    test2 = RpcAlert.new(RpcAlert.WARNING_MIN, "Test2", now2)
    change = value_client.wait_for_change("test/device/.device/alerts")
    await device.change_alerts(add=test2, rem=test1)
    alerts = await change
    assert len(alerts) == 1
    assert alerts[0] == test2.value

    assert list(device.alerts) == [test2]


async def test_no_alerts(client, device):
    """Check that you alerts can't be accessed if they are disabled."""
    device.DEVICE_ALERTS = False
    assert await client.call("test/device/.device", "ls") == []



"""Implementation of RpcMessage."""

from __future__ import annotations

import collections.abc
import enum
import time
import typing

from .chainpack import ChainPackWriter
from .cpon import CponWriter
from .path import SHVPath
from .rpcaccess import RpcAccess
from .rpcerrors import RpcError, RpcErrorCode
from .value import (
    SHVIMap,
    SHVType,
    is_shvlist,
    shvmeta_eq,
)


class RpcMessage:
    """Single SHV RPC message representation.

    :param rpc_val: The value of the received and unpacked SHV RPC message or
      ``None`` in case you want to create a new message instead.
    """

    _last_request_id: typing.ClassVar[int] = 0
    _last_request_id_rollover: typing.ClassVar[float] = time.monotonic()

    @classmethod
    def next_request_id(cls) -> int:
        """Provide unique request identifier.

        The identifier won't repeat for this application in a reasonable time as
        it is just simple counter that wrap after a long time.

        In most cases we could get away with not wrapping the counter at all
        because it is unlikely that we would run out of common implementation
        limit of 64 bits (this is limit in other implementations not Python).
        But further messages would be increased in size. Instead we choose to
        rollover counter every now often (every 15 minutes). That is because of
        a reasonable expectation that request ID is used right away and call
        timeout won't be in minutes.
        """
        tmono = time.monotonic()
        if cls._last_request_id_rollover + (15 * 60) < tmono:
            cls._last_request_id = 0
            cls._last_request_id_rollover = tmono
        cls._last_request_id += 1
        return cls._last_request_id

    def __init__(self, rpc_val: SHVIMap | None = None) -> None:
        if rpc_val is None:
            rpc_val = SHVIMap()
        self.value: SHVIMap = rpc_val

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RpcMessage) and shvmeta_eq(self.value, other.value)

    def __repr__(self) -> str:
        return f"<RpcMessage {self.value.meta!r}: {self.value!r}>"

    class Tag(enum.IntEnum):
        """Tags in Meta for RPC message."""

        META_TYPE_ID = 1
        META_TYPE_NAMESPACE_ID = 2
        REQUEST_ID = 8
        SHV_PATH = 9
        METHOD = 10
        SIGNAL = 10
        CALLER_IDS = 11
        REV_CALLER_IDS = 13
        ACCESS = 14
        USER_ID = 16
        ACCESS_LEVEL = 17
        SOURCE = 19
        REPEAT = 20

    class Key(enum.IntEnum):
        """Keys in the toplevel IMap of the RPC message."""

        PARAM = 1
        RESULT = 2
        ERROR = 3
        DELAY = 4
        ABORT = 5

    class Type(enum.Enum):
        """The message type definition.

        This is defined to allow message differenciate in ``match``. It is more
        specific than SHV standard to allow easier implementation of different
        messages handling.
        """

        REQUEST = enum.auto()
        REQUEST_ABORT = enum.auto()
        RESPONSE = enum.auto()
        RESPONSE_DELAY = enum.auto()
        RESPONSE_ERROR = enum.auto()
        SIGNAL = enum.auto()

    def is_valid(self) -> bool:
        """Check if message is valid RPC message."""
        return (
            isinstance(self.value, SHVIMap)  # Must be IMap
            and len(self.value) <= 1  # Only at most one Key is allowed
            and self.type is not None
            and self.value.meta.get(self.Tag.META_TYPE_ID, 1) == 1
            and self.value.meta.get(self.Tag.META_TYPE_NAMESPACE_ID, 0) == 0
            and isinstance(self._request_id, int | None)
            and isinstance(self._path, str)
            and isinstance(self._signal_name, str)  # Also covers method
            and (
                isinstance(self._caller_ids, int | None)
                or (
                    is_shvlist(self._caller_ids)
                    and all(isinstance(v, int) for v in self._caller_ids)
                )
            )
            and isinstance(self._access, str)
            and isinstance(self._user_id, str | None)
            and isinstance(self._access_level, int | None)
            and isinstance(self._source, str)
            and isinstance(self._repeat, bool)
            # TODO check value content
        )

    @property
    def type(self) -> Type | None:
        """The message type or ``None`` if unknown."""
        if self.Tag.REQUEST_ID in self.value.meta:
            if self.Tag.METHOD in self.value.meta:
                if self.Key.ABORT in self.value:
                    return self.Type.REQUEST_ABORT
                if not self.value or self.Key.PARAM in self.value:
                    return self.Type.REQUEST
            elif self.Key.ERROR in self.value:
                return self.Type.RESPONSE_ERROR
            elif self.Key.DELAY in self.value:
                return self.Type.RESPONSE_DELAY
            elif not self.value or self.Key.RESULT in self.value:
                return self.Type.RESPONSE
        elif not self.value or self.Key.PARAM in self.value:
            return self.Type.SIGNAL
        return None

    def make_response(self, result: SHVType | RpcError = None) -> RpcMessage:
        """Create new message that is response to this one.

        :param result: The result value to be set in the response or
          :py:class:`RpcError` to be reported as error result.
        :return: The new message that is response to this one.
        """
        if self.type not in {self.Type.REQUEST, self.Type.REQUEST_ABORT}:
            raise ValueError("Response can be created from request only.")
        resp = RpcMessage()
        resp.request_id = self.request_id
        resp.caller_ids = self.caller_ids
        if isinstance(result, RpcError):
            resp.error = result
        else:
            resp.result = result
        return resp

    def make_response_delay(self, progress: float = 0.0) -> RpcMessage:
        """Create new message that is response delay for this one.

        :param progress: The progress to be reported.
        :return: The new message that is response delay to this one.
        """
        if self.type not in {self.Type.REQUEST, self.Type.REQUEST_ABORT}:
            raise ValueError("Response delay can be created from request only.")
        resp = RpcMessage()
        resp.request_id = self.request_id
        resp.caller_ids = self.caller_ids
        resp.delay = progress
        return resp

    def make_abort(self, abort: bool) -> RpcMessage:
        """Create abort message for the request."""
        if self.type is not self.Type.REQUEST:
            raise ValueError("Abort request can be created from request only.")
        res = RpcMessage()
        res.request_id = self.request_id
        res.caller_ids = self.caller_ids
        res.method = self.method
        res.path = self.path
        res.abort = abort
        return res

    @property
    def _request_id(self) -> SHVType:
        return self.value.meta.get(self.Tag.REQUEST_ID)

    @property
    def request_id(self) -> int:
        """Request identifier of this message."""
        res = self._request_id
        if not isinstance(res, int):
            raise ValueError(f"Invalid RequestId type: {type(res)}")
        return res

    @request_id.setter
    def request_id(self, rqid: int | None) -> None:
        """Set given request identifier to this message."""
        if rqid is None:
            self.value.meta.pop(self.Tag.REQUEST_ID, None)
        else:
            self.value.meta[self.Tag.REQUEST_ID] = rqid

    def new_request_id(self) -> int:
        """Set new request ID.

        :return: The new request ID.
        """
        self.request_id = self.next_request_id()
        return self.request_id

    @property
    def _path(self) -> SHVType:
        return self.value.meta.get(self.Tag.SHV_PATH, "")

    @property
    def path(self) -> str:
        """SHV path specified for this message or empty string."""
        res = self._path
        if not isinstance(res, str):
            raise ValueError(f"Invalid ShvPath type: {type(res)}")
        return res

    @path.setter
    def path(self, path: str) -> None:
        """Set given path as SHV path for this message."""
        if path:
            self.value.meta[self.Tag.SHV_PATH] = path
        else:
            self.value.meta.pop(self.Tag.SHV_PATH, None)

    @property
    def shvpath(self) -> SHVPath:
        """SHV path specified for this message as :class:`SHVPath`."""
        return SHVPath(self.path)

    @shvpath.setter
    def shvpath(self, path: SHVPath) -> None:
        """Set given :class:`SHVPath` as SHV path for this message."""
        self.path = str(path)

    @property
    def method(self) -> str:
        """SHV method name for this message."""
        res = self.value.meta.get(self.Tag.METHOD)
        if not isinstance(res, str):
            raise ValueError(f"Invalid Method type: {type(res)}")
        return res

    @method.setter
    def method(self, method: str) -> None:
        """Set SHV method name for this message."""
        if method:
            self.value.meta[self.Tag.METHOD] = method
        else:
            self.value.meta.pop(self.Tag.METHOD, None)

    @property
    def _signal_name(self) -> SHVType:
        return self.value.meta.get(self.Tag.METHOD, "chng")

    @property
    def signal_name(self) -> str:
        """SHV signal name for this message."""
        res = self._signal_name
        if not isinstance(res, str):
            raise ValueError(f"Invalid Signal type: {type(res)}")
        return res

    @signal_name.setter
    def signal_name(self, signal: str) -> None:
        """Set SHV signal name for this message."""
        # Note: we always set it because old implementations were dropping
        # messages without method name.
        self.value.meta[self.Tag.SIGNAL] = signal

    @property
    def _source(self) -> SHVType:
        return self.value.meta.get(self.Tag.SOURCE, "get")

    @property
    def source(self) -> str:
        """SHV signal source method name for this message."""
        res = self._source
        if not isinstance(res, str):
            raise ValueError(f"Invalid Source type: {type(res)}")
        return res

    @source.setter
    def source(self, source: str) -> None:
        """Set SHV signal source method name for this message."""
        if source and source != "get":
            self.value.meta[self.Tag.SOURCE] = source
        else:
            self.value.meta.pop(self.Tag.SOURCE, None)

    @property
    def _caller_ids(self) -> SHVType:
        return self.value.meta.get(self.Tag.CALLER_IDS, None)

    @property
    def caller_ids(self) -> collections.abc.Sequence[int]:
        """Caller identifiers associated with this message."""
        res = self._caller_ids
        if res is None:
            return []
        if isinstance(res, int):
            return [res]
        if not is_shvlist(res) or any(not isinstance(v, int) for v in res):
            raise ValueError(f"Invalid CallerIds type: {type(res)}")
        # Mypy doesn't understand the any check and thus we have to type cast
        return typing.cast(collections.abc.Sequence[int], res)

    @caller_ids.setter
    def caller_ids(self, cids: collections.abc.Sequence[int]) -> None:
        """Set caller identifiers associated with this message."""
        if not cids:
            self.value.meta.pop(self.Tag.CALLER_IDS, None)
        elif len(cids) == 1:
            self.value.meta[self.Tag.CALLER_IDS] = cids[0]
        else:
            self.value.meta[self.Tag.CALLER_IDS] = cids

    @property
    def _access(self) -> SHVType:
        return self.value.meta.get(self.Tag.ACCESS, "")

    @property
    def access(self) -> collections.abc.Sequence[str]:
        """Granted access sequence."""
        res = self._access
        if not isinstance(res, str):
            raise ValueError(f"Invalid access type: {type(res)}")
        if res:
            return res.split(",")
        return []

    @access.setter
    def access(self, access: collections.abc.Sequence[str]) -> None:
        """Set granted access sequence."""
        if access:
            self.value.meta[self.Tag.ACCESS] = access
        else:
            self.value.meta.pop(self.Tag.ACCESS, None)

    @property
    def _access_level(self) -> SHVType:
        return self.value.meta.get(self.Tag.ACCESS_LEVEL)

    @property
    def rpc_access(self) -> RpcAccess | None:
        """Access level as :class:`shv.RpcAccess`."""
        if (level := self._access_level) is not None:
            if not isinstance(level, int):
                raise ValueError(f"Invalid AccessLevel type: {type(level)}")
            return RpcAccess(level)
        m = RpcAccess.strmap()
        for access in self.access:
            if access in m:
                return m[access]
        return None

    @rpc_access.setter
    def rpc_access(self, access: RpcAccess | None) -> None:
        """Set access level with :class:`shv.RpcAccess`."""
        if access is not None:
            self.value.meta[self.Tag.ACCESS] = RpcAccess.tostr(access)
            self.value.meta[self.Tag.ACCESS_LEVEL] = access.value
        else:
            self.value.meta.pop(self.Tag.ACCESS, None)

    @property
    def _user_id(self) -> SHVType:
        res = self.value.meta.get(self.Tag.USER_ID, None)
        if isinstance(res, dict):  # Note: backward compatibility
            res = f"{res.get('brokerId')}:{res.get('shvUser')}"
        return res

    @property
    def user_id(self) -> str | None:
        """User's ID carried by message."""
        res = self._user_id
        if not isinstance(res, str | None):
            raise ValueError(f"Invalid UserId type: {type(res)}")
        return res

    @user_id.setter
    def user_id(self, value: str | None) -> None:
        """Set User's ID."""
        if value is not None:
            self.value.meta[self.Tag.USER_ID] = value
        else:
            self.value.meta.pop(self.Tag.USER_ID, None)

    @property
    def _repeat(self) -> SHVType:
        return self.value.meta.get(self.Tag.REPEAT, False)

    @property
    def repeat(self) -> bool:
        """Signal is possibly a repeat of some previous signal."""
        res = self._repeat
        if not isinstance(res, bool):
            raise ValueError(f"Invalid Repeat type: {type(res)}")
        return res

    @repeat.setter
    def repeat(self, value: bool | None) -> None:
        """Set repeat."""
        if value is not None:
            self.value.meta[self.Tag.REPEAT] = value
        else:
            self.value.meta.pop(self.Tag.REPEAT, None)

    @property
    def param(self) -> SHVType:
        """SHV parameters for the method call.

        Usable only for :py:data:`Type.REQUEST` and :py:data:`Type.SIGNAL`.
        """
        return self.value.get(self.Key.PARAM, None)

    @param.setter
    def param(self, param: SHVType) -> None:
        """Set SHV parameters for this method call."""
        if param is None:
            self.value.pop(self.Key.PARAM, None)
        else:
            self.value[self.Key.PARAM] = param

    @property
    def abort(self) -> bool:
        """The delay progress for the :py:data:`Type.REQUEST_ABORT`."""
        res = self.value.get(self.Key.ABORT)
        if not isinstance(res, bool):
            raise ValueError(f"Invalid Abort: {res!r}")
        return res

    @abort.setter
    def abort(self, abort: bool | None) -> None:
        """Set SHV Request Delay progress."""
        if abort is None:
            self.value.pop(self.Key.ABORT, None)
        else:
            self.value[self.Key.ABORT] = abort

    @property
    def result(self) -> SHVType:
        """SHV method call result.

        Usable only for :py:data:`Type.RESPONSE`.
        """
        return self.value.get(self.Key.RESULT, None)

    @result.setter
    def result(self, result: SHVType) -> None:
        """Set SHV method call result."""
        if result is None:
            self.value.pop(self.Key.RESULT, None)
        else:
            self.value[self.Key.RESULT] = result

    @property
    def error(self) -> RpcError:
        """SHV method call error.

        Usable only for :py:data:`Type.ERROR`.
        """
        return RpcError.from_shv(self.value.get(self.Key.ERROR))

    @error.setter
    def error(self, error: RpcError | None) -> None:
        """Set SHV method call error."""
        if error is None or error.error_code == RpcErrorCode.NO_ERROR:
            self.value.pop(self.Key.ERROR, None)
        else:
            self.value[self.Key.ERROR] = error.to_shv()

    @property
    def delay(self) -> float:
        """The delay progress for the :py:data:`Type.RESPONSE_DELAY`."""
        res = self.value.get(self.Key.DELAY)
        if not isinstance(res, float):
            raise ValueError(f"Invalid Delay: {res!r}")
        return res

    @delay.setter
    def delay(self, progress: float | None) -> None:
        """Set SHV Request Delay progress."""
        if progress is None:
            self.value.pop(self.Key.DELAY, None)
        else:
            self.value[self.Key.DELAY] = progress

    def to_string(self) -> str:
        """Convert message to CPON and return it as string."""
        return self.to_cpon().decode("utf-8")

    def to_cpon(self) -> bytes:
        """Convert message to Cpon."""
        return CponWriter.pack(self.value)

    def to_chainpack(self) -> bytes:
        """Convert message to Chainpack."""
        return ChainPackWriter.pack(self.value)

    @classmethod
    def request(
        cls,
        path: str | SHVPath,
        method: str,
        param: SHVType = None,
        rid: int | None = None,
        cids: collections.abc.Sequence[int] = tuple(),
        user_id: str | None = None,
    ) -> RpcMessage:
        """Create request message.

        :param path: SHV path for signal.
        :param method: method name for signal.
        :param param: Parameters passed to the method.
        :param rid: Request identifier for this message. It is automatically
          assigned if ``None`` is passed.
        :param cids: The caller IDs. This can be used with methods using a
          unique CallerIds sequence to establish session.
        :param user_id: User's ID to be caried with request message.
        """
        res = cls()
        res.request_id = rid or cls.next_request_id()
        res.caller_ids = cids
        res.method = method
        res.path = str(path)
        res.param = param
        res.user_id = user_id
        return res

    @classmethod
    def signal(
        cls,
        path: str | SHVPath,
        name: str = "chng",
        source: str = "get",
        value: SHVType = None,
        access: RpcAccess = RpcAccess.READ,
        user_id: str | None = None,
    ) -> RpcMessage:
        """Create signal message.

        :param path: SHV path for signal.
        :param name: Name of the signal.
        :param source: Name of the method this signal is associated with.
        :param value: Value to be sent in the message.
        :param access: Minimal access level needed to get this signal.
        :param user_id: User ID associated with this signal.
        """
        res = cls()
        res.signal_name = name
        res.source = source
        res.path = str(path)
        res.param = value
        res.rpc_access = access
        res.user_id = user_id
        return res

    @classmethod
    def lsmod(
        cls, path: str | SHVPath, nodes: collections.abc.Mapping[str, bool]
    ) -> RpcMessage:
        """Create ``lsmod`` signal message.

        This provides creation of "lsmod" signal message that must be used when
        you are changing the nodes tree to signal clients about that. The
        argument specifies top level nodes added or removed (based on the
        mapping value).

        :param path: SHV path to the valid node which children were added or
          removed.
        :param nodes: Map where key is node name of the node that is top level
          node, that was either added (for value ``True``) or removed (for value
          ``False``).
        """
        return cls.signal(path, "lsmod", "ls", nodes, RpcAccess.BROWSE)



"""The SHV RPC type key-struct."""

from __future__ import annotations

import collections.abc
import typing

from .. import SHVMapType, SHVType
from .base import RpcType


class RpcTypeKeyStruct(RpcType, collections.abc.Mapping[str, RpcType]):
    """The KeyStruct type representation."""

    def __init__(self, items: collections.abc.Mapping[str, RpcType]) -> None:
        self._items = items

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RpcTypeKeyStruct) and self._items == other._items

    def __str__(self) -> str:
        return f"{{{','.join(f'{tp}:{key}' for key, tp in self._items.items())}}}"

    def __getitem__(self, key: str) -> RpcType:
        return self._items[key]

    def __iter__(self) -> collections.abc.Iterator[str]:
        return iter(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def validate(  # noqa: D102
        self, value: SHVType
    ) -> typing.TypeGuard[collections.abc.Mapping[str, SHVType]]:
        return (
            isinstance(value, collections.abc.Mapping)
            and all(k in self._items for k in value)
            and all(
                v.validate(typing.cast(SHVMapType, value).get(k))
                for k, v in self._items.items()
            )
        )



"""Verify implementation of RPC file helpers."""

import collections.abc
import contextlib
import datetime
import pathlib
import typing

import pytest

from shv import (
    FileProviderRW,
    RpcAccess,
    RpcDir,
    RpcFile,
    RpcFileStat,
    SHVBase,
    SHVClient,
    SHVType,
)


class FilesProvider(SHVClient):
    APP_NAME = "testdev"

    def __init__(
        self,
        *args: typing.Any,  # noqa ANNN401
        path: pathlib.Path,
        **kwargs: typing.Any,  # noqa ANNN401
    ) -> None:
        self.path = path
        super().__init__(*args, **kwargs)  # notype

    def _ls(self, path: str) -> collections.abc.Iterator[str]:
        yield from super()._ls(path)
        if not path:
            yield "files"
        elif path == "files" or path.startswith("files/"):
            pth = self.path / path[6:]
            if pth.is_dir():
                yield from sorted(spth.name for spth in pth.iterdir())

    def _dir(self, path: str) -> collections.abc.Iterator[RpcDir]:
        yield from super()._dir(path)
        if path.startswith("files/"):
            pth = self.path / path[6:]
            if pth.is_file():
                yield from FileProviderRW.dir()

    async def _method_call(self, request: SHVBase.Request) -> SHVType:
        if request.path.startswith("files/"):
            pth = self.path / request.path[6:]
            if pth.is_file():
                with contextlib.suppress(NotImplementedError):
                    # TODO use also other providers
                    return await FileProviderRW.method_call(pth, request)
        return await super()._method_call(request)


@pytest.fixture(name="device")
async def fixture_device(shvbroker, url_test_device, tmp_path):
    """Run FilesProvider and provide instance to access it."""
    device = await FilesProvider.connect(url_test_device, path=tmp_path)
    yield device
    await device.disconnect()


@pytest.fixture(name="seed_tmp")
def fixture_seed_tmp(tmp_path):
    """Create some dummy files in tmp_path."""
    (tmp_path / "bin").write_bytes(b"abc")
    (tmp_path / "lorem.txt").write_text(
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
    )
    (tmp_path / "sub" / "subsub").mkdir(parents=True)


@pytest.mark.parametrize(
    "path,res",
    (
        ("test/device", [".app", "files"]),
        ("test/device/files", ["bin", "lorem.txt", "sub"]),
        ("test/device/files/bin", []),
        ("test/device/files/lorem.txt", []),
        ("test/device/files/sub", ["subsub"]),
        ("test/device/files/sub/subsub", []),
    ),
)
async def test_ls(client, device, seed_tmp, path, res):
    assert await client.ls(path) == res


@pytest.mark.parametrize(
    "path,res",
    (
        (
            "test/device/files/lorem.txt",
            [
                RpcDir.stddir(),
                RpcDir.stdls(),
                RpcDir.getter("stat", "n", "!stat", RpcAccess.READ),
                RpcDir.getter("size", "n", "i(0,)", RpcAccess.READ),
                RpcDir(
                    "crc",
                    param="[i(0,):offset,i(0,)|n:size]|n",
                    result="u(>32)",
                    access=RpcAccess.READ,
                ),
                RpcDir(
                    "sha1",
                    param="[i(0,):offset,i(0,)|n:size]|n",
                    result="b(20)",
                    access=RpcAccess.READ,
                ),
                RpcDir(
                    "read",
                    RpcDir.Flag.LARGE_RESULT_HINT,
                    "[i(0,):offset,i(0,)size]",
                    "b",
                    RpcAccess.READ,
                ),
                RpcDir(
                    "write",
                    param="[i(0,):offset,b:data]",
                    result="n",
                    access=RpcAccess.WRITE,
                ),
                RpcDir("truncate", param="i(0,)", result="n", access=RpcAccess.WRITE),
                RpcDir("append", param="b", result="n", access=RpcAccess.WRITE),
            ],
        ),
    ),
)
async def test_dir(client, device, seed_tmp, path, res):
    assert await client.dir(path) == res


@pytest.fixture(name="file_bin")
def fixture_file_bin(client, device, seed_tmp):
    return RpcFile(client, "test/device/files/bin", RpcFile.Flag(0))


def test_file_bin_path(file_bin):
    assert file_bin.path == "test/device/files/bin"


async def test_file_bin_readable(file_bin):
    assert await file_bin.readable()


async def test_file_bin_writable(file_bin):
    assert await file_bin.writable()


async def test_file_bin_resizable(file_bin):
    assert await file_bin.resizable()


async def test_file_bin_stat(file_bin, tmp_path):
    stat = (tmp_path / "bin").stat()
    access = datetime.datetime.fromtimestamp(stat.st_atime, tz=datetime.UTC)
    access = access.replace(microsecond=(access.microsecond // 1000) * 1000)
    mod = datetime.datetime.fromtimestamp(stat.st_mtime, tz=datetime.UTC)
    mod = access.replace(microsecond=(mod.microsecond // 1000) * 1000)
    assert await file_bin.stat() == RpcFileStat(
        stat.st_size, stat.st_blksize, access, mod
    )


async def test_file_bin_size(client, device, seed_tmp):
    assert await client.call("test/device/files/bin", "size") == 3


@pytest.mark.parametrize(
    "offset,size,res",
    (
        (0, None, 891568578),
        (1, None, 3265866552),
        (0, 1, 3904355907),
    ),
)
async def test_file_bin_crc32(file_bin, offset, size, res):
    assert await file_bin.crc32(offset, size) == res


@pytest.mark.parametrize(
    "offset,size,res",
    (
        (0, None, b"\xa9\x99>6G\x06\x81j\xba>%qxP\xc2l\x9c\xd0\xd8\x9d"),
        (1, None, b"[%\x05\x03\x9a\xc5\xaf\x9e\x19\x7f]\xad\x04\x119\x06\xa9\xcf\x9a*"),
        (0, 1, b"\x86\xf7\xe47\xfa\xa5\xa7\xfc\xe1]\x1d\xdc\xb9\xea\xea\xea7vg\xb8"),
    ),
)
async def test_file_bin_sha1(file_bin, offset, size, res):
    assert await file_bin.sha1(offset, size) == res


async def test_file_bin_read(file_bin):
    assert file_bin.offset == 0
    assert await file_bin.read() == b"abc"
    assert file_bin.offset == 3


async def test_file_bin_read_short(file_bin):
    assert file_bin.offset == 0
    assert await file_bin.read(2) == b"ab"
    assert file_bin.offset == 2
    assert await file_bin.read(2) == b"c"
    assert file_bin.offset == 3


async def test_file_bin_write(file_bin, tmp_path):
    file_bin.offset = 3
    await file_bin.write(b"def")
    assert file_bin.offset == 6
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b"abcdef"


async def test_file_bin_flush(file_bin, tmp_path):
    # This does nothing so just call it that it won't fail
    await file_bin.flush()


async def test_file_bin_truncate(file_bin, tmp_path):
    await file_bin.truncate()
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b""


@pytest.fixture(name="file_append")
def fixture_file_append(client, device, seed_tmp):
    return RpcFile(client, "test/device/files/bin", RpcFile.Flag.APPEND)


async def test_file_bin_append(file_append, tmp_path):
    file_append.offset = 0
    await file_append.write(b"def")
    file_append.offset = 0
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b"abcdef"


@pytest.fixture(name="file_buffered")
def fixture_file_buffered(client, device, seed_tmp):
    return RpcFile(client, "test/device/files/bin", RpcFile.Flag.BUFFERED)


async def test_file_buffered_read(file_buffered, tmp_path):
    assert file_buffered.offset == 0
    assert await file_buffered.read(1) == b"a"
    assert file_buffered.offset == 1
    assert await file_buffered.read(1) == b"b"
    assert file_buffered.offset == 2
    assert await file_buffered.read(2) == b"c"
    assert file_buffered.offset == 3
    assert await file_buffered.read(2) == b""
    assert file_buffered.offset == 3
    with (tmp_path / "bin").open("ab") as file:
        file.write(b"def")
    assert await file_buffered.read(2) == b"de"
    assert file_buffered.offset == 5


async def test_file_buffered_write(file_buffered, tmp_path):
    file_buffered.offset = await file_buffered.size()
    await file_buffered.write(b"de")
    await file_buffered.write(b"f")
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b"abc"
    await file_buffered.flush()
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b"abcdef"

    await file_buffered.write(b"gh")
    file_buffered.offset = 1
    await file_buffered.write(b"d")
    with (tmp_path / "bin").open("rb") as file:
        assert file.read() == b"abcdefgh"



"""The base for the various high level SHV RPC interfaces."""

from __future__ import annotations

import asyncio
import collections
import collections.abc
import contextlib
import datetime
import functools
import logging
import time

from .__version__ import VERSION
from .path import SHVPath
from .rpcaccess import RpcAccess
from .rpcdir import RpcDir
from .rpcerrors import (
    RpcError,
    RpcInvalidParamError,
    RpcMethodCallExceptionError,
    RpcMethodNotFoundError,
    RpcRequestInvalidError,
    RpcTryAgainLaterError,
    RpcUserIDRequiredError,
)
from .rpcmessage import RpcMessage
from .rpcri import rpcri_match
from .rpctransport import RpcClient
from .shvversion import SHV_VERSION_MAJOR, SHV_VERSION_MINOR
from .value import SHVType, is_shvbool, is_shvnull

logger = logging.getLogger(__name__)


class SHVBase:
    """SHV RPC object API base class.

    You want to use this if you plan to implement your own specific RPC message
    handler but in most cases you might want to use :class:`shv.SHVClient` or
    :class:`shv.SHVDevice` instead.

    Messages are handled in an asyncio loop and based on the type of the message
    the different operation is performed.

    :param client: The RPC client instance to wrap and manage.
    :param call_timeout: Timeout in seconds before call is abandoned. The
      default is no timeout and thus infinite waiting. This is handy if you
      don't want to wrap call calls in your code with :func:`asyncio.timeout`
      and does exactly the same thing.
    :param call_query_timeout: Timeout in seconds before query is used to check
      the request status. The shorter time will cause faster message lost
      detection while the longer
    :param call_retry_timeout: Timeout in seconds when the request is sent
      again if there is no response received from the device.
    :param user_id: The default user ID to be used.
    :param peer_shv_version: The assumed SHV version of the peer. In default
      the exact version is detected but sometimes it is desirable to skip this
      check of enforce a different version for testing purposes.
    """

    IDLE_TIMEOUT: float = 180
    """Number of seconds before we are disconnected from the broker automatically."""

    APP_NAME: str = "pyshv"
    """Name of the application reported to the SHV.

    You can change this value in child class to report a more accurate
    application name.
    """
    APP_VERSION: str = VERSION
    """Version of the application reported to the SHV.

    You should change this value in child class to report a correct number.
    """

    def __init__(
        self,
        client: RpcClient,
        call_query_timeout: float = 1.0,
        call_retry_timeout: float = 60.0,
        user_id: str = "",
        peer_shv_version: tuple[int, int] | None = None,
    ) -> None:
        self.client = client
        """The underlining RPC client instance.

        **Do not send messages by directly accessing this property. You must use
        implementations provided by this class!**
        """
        self.task = asyncio.create_task(self._loop())
        """Task running the message handling receive loop."""
        self.user_id = user_id
        """The default user ID provided if method should be requested with it."""
        self.call_query_timeout = call_query_timeout
        """Timeout in seconds before :meth:`call` queries the state of the request."""
        self.call_retry_timeout = call_retry_timeout
        """Timeout in seconds before :meth:`call` send request again.

        Be aware how this interacts with :attr:`SHVBase.call_query_timeout`.
        This should be longer to ensure that we try to query the request before
        it is submitted as the whole again.

        This timeout is applied only if we get no response for given amount of
        time. Any response on query on voluntary one extends this timeout.
        """
        self._requests: dict[tuple[int, ...], tuple[asyncio.Task, SHVBase.Request]] = {}
        self._responses: dict[int, asyncio.Queue[RpcMessage | None]] = {}
        self._peer_shv_version = peer_shv_version
        self.__initial_peer_shv_version = peer_shv_version
        self.__send_semaphore = asyncio.Semaphore()
        self.__send_queue: collections.deque[tuple[RpcMessage, asyncio.Future]]
        self.__send_queue = collections.deque()

    async def disconnect(self) -> None:
        """Disconnect an existing connection.

        The call to the disconnect when client is not connected is silently
        ignored.

        Await blocks until disconnect is completed. You can use
        ``x.client.disconnect()`` to only initialize disconnect without waiting
        for it to take an effect.
        """
        self.client.disconnect()
        await self.task

    async def _loop(self) -> None:
        """Loop run in asyncio task to receive messages."""
        send_task = asyncio.create_task(self._send_loop())
        try:
            with contextlib.suppress(EOFError):
                while msg := await self.client.receive():
                    if msg is RpcClient.Control.RESET:
                        self._reset()
                    elif msg.is_valid():
                        await self._message(msg)
                    else:
                        logger.info("%s: Dropped invalid message: %s", self.client, msg)
        finally:
            send_task.cancel()
            tasks = [send_task]
            for request_task, _ in self._requests.values():
                request_task.cancel()
                tasks.append(request_task)
            res = await asyncio.gather(*tasks, return_exceptions=True)
            if excs := [
                v
                for v in res
                if isinstance(v, BaseException)
                and not isinstance(v, asyncio.CancelledError)
            ]:
                if len(excs) == 1:
                    raise excs[0]
                raise BaseExceptionGroup("Collected errors from SHVBase", excs)

    def _reset(self) -> None:
        """Handle peer's reset request."""
        logger.info("%s: Doing reset", self.client)
        self._peer_shv_version = self.__initial_peer_shv_version
        for request_task, _ in self._requests.values():
            request_task.cancel()
        for queue in self._responses.values():
            with contextlib.suppress(asyncio.QueueFull):
                queue.put_nowait(None)

    async def reset(self) -> None:
        """Reset the client and connection.

        This calls :meth:`RpcClient.reset` as well as reset of the internal
        state.
        """
        await self.client.reset()
        self._reset()

    async def call(
        self,
        path: str | SHVPath,
        method: str,
        param: SHVType = None,
        user_id: str | None = None,
        query_timeout: float | None = None,
        retry_timeout: float | None = None,
        progress: collections.abc.Callable[
            [float | None], collections.abc.Awaitable[None] | None
        ]
        | None = None,
    ) -> SHVType:
        """Call given method on given path with given parameter.

        Note that this is coroutine and thus it is up to you if you await it or
        use asyncio tasks. You can also freely use :func:`asyncio.timeout` to
        timeout the call.

        The delivery of the messages is not ensure in SHV network (they can be
        dropped due to multiple reasons without informing the source of the
        message) and thus this method can send the request message multiple
        times if it doesn't receive an appropriate response.

        :param path: SHV path method is associated with.
        :param method: SHV method name to be called.
        :param param: Parameter passed to the called method.
        :param user_id: UserID added to the method call request. This is required
          by some RPC methods to identify the user and they will respond with
          :class:`RpcUserIDRequiredError` if it is missing. This is caught by
          this method and request will be attempted again with User ID ``""``.
          If you know that method needs User ID then you can prevent this round
          trip by setting this argument to ``""``. On the other hand sending
          all requests with User ID wastes with bandwidth. In general usage you
          should use :attr:`SHVBase.user_id` instead of just ``""`` to use the
          object default.
        :param query_timeout: Override :attr:`SHVBase.call_query_timeout` just
          for this one call.
        :param retry_timeout: Override :attr:`SHVBase.call_retry_timeout` just
          for this one call.
        :param progress: An optional callback function that is called to report
          on progress of the call. It can be either function of coroutine. The
          argument passed is the float between 0 and 1 providing the progress
          signalization or ``None`` to signal that
          :class:`RpcTryAgainLaterError` was received. If you need to pass
          additional arguments to the function then use
          :func:`functools.partial`.
        :return: Return value on successful method call.
        :raise RpcError: The call result in error that is propagated by raising
          `RpcError` or its children based on the failure.
        :raise EOFError: when client disconnected and thus request can't be
          sent or response received.
        """
        query_timeout = (
            self.call_query_timeout if query_timeout is None else query_timeout
        )
        retry_timeout = (
            self.call_retry_timeout if retry_timeout is None else retry_timeout
        )

        async def callback_progress(value: float | None) -> None:
            if progress is not None:
                if (ares := progress(value)) is not None:
                    await ares

        request = RpcMessage.request(path, method, param, user_id=user_id)
        assert request.request_id not in self._responses
        queue = self._responses[request.request_id] = asyncio.Queue()
        try:
            while True:
                await self._send(request)
                last = time.monotonic()
                while True:
                    try:
                        tm = min(query_timeout, retry_timeout + last - time.monotonic())
                        async with asyncio.timeout(tm):
                            msg = await queue.get()
                        if msg is None:  # The communication reset
                            break  # We need to send request again
                        match msg.type:
                            case RpcMessage.Type.RESPONSE:
                                return msg.result
                            case RpcMessage.Type.RESPONSE_ERROR:
                                match msg.error:
                                    case RpcUserIDRequiredError():
                                        request.user_id = self.user_id
                                        rid = request.new_request_id()
                                        self._responses[rid] = queue
                                    case RpcRequestInvalidError():
                                        self._responses[request.request_id] = queue
                                    case RpcTryAgainLaterError():
                                        await callback_progress(None)
                                        rid = request.new_request_id()
                                        self._responses[rid] = queue
                                        tm = retry_timeout + last - time.monotonic()
                                        if tm > 0:
                                            await asyncio.sleep(tm)
                                    case error:
                                        raise error
                                break
                            case RpcMessage.Type.RESPONSE_DELAY:
                                await callback_progress(msg.delay)
                                last = time.monotonic()
                            case _:  # pragma: no cover
                                raise AssertionError(f"Invalid message: {msg}")
                    except TimeoutError:
                        if time.monotonic() - last > retry_timeout:
                            break
                        await self._send(request.make_abort(False))
        finally:
            if request.request_id in self._responses:
                if self.client.connected:
                    # TODO possibly put to the task
                    await self._send(request.make_abort(True))
                del self._responses[request.request_id]

    async def ping(self) -> None:
        """Ping the peer to check the connection."""
        await self.call(
            ".app" if await self.peer_shv_version() >= (3, 0) else ".broker/app", "ping"
        )

    async def ls(self, path: str | SHVPath) -> list[str]:
        """List child nodes of the node on the specified path.

        :param path: SHV path to the node we want children to be listed for.
        :return: list of child nodes.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "ls")
        if not isinstance(res, list):  # pragma: no cover
            raise RpcMethodCallExceptionError(f"Invalid result returned: {res!r}")
        return res

    async def ls_has_child(self, path: str | SHVPath, name: str) -> bool:
        """Use ``ls`` method to check for child.

        :param path: SHV path to the node with possible child of given name.
        :param name: Name of the child node.
        :return: ``True`` if there is such child and ``False`` if not.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "ls", name)
        if not isinstance(res, bool):  # pragma: no cover
            raise RpcMethodCallExceptionError(f"Invalid result returned: {res!r}")
        return res

    async def dir(self, path: str | SHVPath, details: bool = False) -> list[RpcDir]:
        """List methods associated with node on the specified path.

        :param path: SHV path to the node we want methods to be listed for.
        :param details: If detailed listing should be performed instead of standard one.
        :return: list of the node's methods.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "dir", True if details else None)
        if isinstance(res, list):  # pragma: no cover
            return [RpcDir.from_shv(m) for m in res]
        raise RpcMethodCallExceptionError(f"Invalid result returned: {res!r}")

    async def dir_exists(self, path: str | SHVPath, name: str) -> bool:
        """Check if method exists using ``dir`` method.

        :param path: SHV path to the node we want methods to be listed for.
        :param name: Name of the method which existence should be checked.
        :return: ``True`` if method exists and ``False`` otherwise.
        :raise RpcMethodNotFoundError: when there is no such path.
        """
        res = await self.call(path, "dir", name)
        # list, null and mapping is backward compatibility
        if not isinstance(
            res, bool | collections.abc.Mapping | collections.abc.Sequence | None
        ):  # pragma: no cover
            raise RpcMethodCallExceptionError(f"Invalid result returned: {res!r}")
        return bool(res)

    async def peer_shv_version(self) -> tuple[int, int]:
        """Get the peer's SHV version."""
        if self._peer_shv_version is None:
            self._peer_shv_version = (0, 0)
            with contextlib.suppress(RpcError):
                major = await self.call(".app", "shvVersionMajor")
                minor = await self.call(".app", "shvVersionMinor")
                if isinstance(major, int) and isinstance(minor, int):
                    self._peer_shv_version = (major, minor)
        return self._peer_shv_version

    async def _send_loop(self) -> None:
        """Loop used to send messages.

        We have a dedicated loop to send messages to identify if there is an
        empty bandwidth. Simply if there are no messages to be sent through
        :meth:`_send` then we can use :meth:`_idle_message` to generate idle
        messages.
        """
        while True:
            await self.__send_semaphore.acquire()
            if self.__send_queue:
                msg, future = self.__send_queue.popleft()
            else:
                res = self._idle_message()
                if res is None:
                    continue
                msg, future = res, None
            try:
                await self.client.send(msg)
            except Exception as exc:
                if future is None:
                    logger.warning("Idle send failed", exc_info=exc)
                else:
                    future.set_exception(exc)
            else:
                if future is not None:
                    future.set_result(None)

    async def _send(self, msg: RpcMessage) -> None:
        """Send message.

        You should be using this method instead of ``self.client.send`` to
        ensure that send can be correctly overwritten and optionally postponed
        or blocked by child implementations.

        This is intentionally marked as protected (accessible only by class and
        its children methods) because generic message must be send only for
        valid paths and methods and that is something only class itself can
        ensure. Your implemntation should provide public methods that protects
        against call with invalid path or method.

        Note that this is coroutine and thus it is up to you if you await it or
        use asyncio tasks.

        :param msg: The message to be send.
        """
        future = asyncio.get_running_loop().create_future()
        self.__send_queue.append((msg, future))
        self.__send_semaphore.release()
        await future

    def _idle_message(self) -> RpcMessage | None:
        """Messages to be sent when there is bandwidth for them.

        In general we just try to propagate all messages but some of the
        messages might not have to be sent if we don't have bandwidth for them
        at the moment. These are commonly messages not directly requested by
        other side such as signals or some of the delayed responses.
        """
        for _, request in self._requests.values():
            if request._progress_dirty:
                request._progress_dirty = False
                return request._msg.make_response_delay(request._progress)
        return None

    def _idle_message_ready(self) -> None:
        """Notify internal implementation that idle message could be generated."""
        self.__send_semaphore.release()

    async def _message(self, msg: RpcMessage) -> None:
        """Handle every received message.

        :param msg: Received message.
        """
        match msg.type:
            case RpcMessage.Type.REQUEST:
                key = (msg.request_id, *msg.caller_ids)
                task = asyncio.create_task(self.__method_call(msg))
                self._requests[key] = (task, self.Request(msg, self))
                task.add_done_callback(functools.partial(self._requests.pop, key))
            case RpcMessage.Type.REQUEST_ABORT:
                await self.__method_call(msg)
            case RpcMessage.Type.RESPONSE | RpcMessage.Type.RESPONSE_ERROR:
                if (queue := self._responses.pop(msg.request_id, None)) is not None:
                    await queue.put(msg)
            case RpcMessage.Type.RESPONSE_DELAY:
                if (queue := self._responses.get(msg.request_id)) is not None:
                    await queue.put(msg)
            case RpcMessage.Type.SIGNAL:
                await self._got_signal(self.Signal(msg))

    async def __method_call(self, msg: RpcMessage) -> None:
        try:
            key = (msg.request_id, *msg.caller_ids)
            if msg.type == RpcMessage.Type.REQUEST:
                resp = msg.make_response()
                try:
                    resp.result = await self._method_call(self._requests[key][1])
                except Exception as exc:
                    resp.error = RpcError.from_exception(exc)
            elif key in self._requests and not msg.abort:
                resp = msg.make_response_delay(self._requests[key][1].progress)
            else:
                resp = msg.make_response()
                if key in self._requests:
                    self._requests[key][0].cancel()
                    resp.error = RpcRequestInvalidError("Request cancelled")
                else:
                    resp.error = RpcRequestInvalidError("No such request")

            with contextlib.suppress(EOFError):  # No need to spam logs on disconnect
                await self._send(resp)
        except Exception as exc:
            logger.warning("%s: Failed to respond: %s", self.client, msg, exc_info=exc)

    async def _method_call(self, request: SHVBase.Request) -> SHVType:
        """Handle request.

        Be aware that :class:`asyncio.CancelledError` can be raised to abort
        the request.

        :param request: SHV RPC request message info.
        :return: result of the method call. To report error you should raise
            :exc:`RpcError`.
        """
        match request.path, request.method:
            case ".app", "shvVersionMajor":
                return SHV_VERSION_MAJOR
            case ".app", "shvVersionMinor":
                return SHV_VERSION_MINOR
            case ".app", "name":
                return self.APP_NAME
            case ".app", "version":
                return self.APP_VERSION
            case ".app", "date":
                return datetime.datetime.now().astimezone()
            case ".app", "ping":
                return None
            case _, "ls":
                return self._method_call_ls(request.path, request.param)
            case _, "dir":
                return self._method_call_dir(request.path, request.param)
        raise RpcMethodNotFoundError(
            f"No such path '{request.path}' or method '{request.method}' or access rights."
        )

    def _method_call_ls(self, path: str, param: SHVType) -> SHVType:
        """Implement ``ls`` method call functionality."""
        if is_shvnull(param):
            res = []
            for v in self._ls(path):
                if v not in res:
                    res.append(v)
            if not res and not self._valid_path(path):
                raise RpcMethodNotFoundError(f"No such node: {path}")
            return res
        if isinstance(param, str):
            return any(v == param for v in self._ls(path))
        raise RpcInvalidParamError("Use Null or String with node name")

    def _ls(self, path: str) -> collections.abc.Iterator[str]:  # noqa: PLR6301
        """Implement ``ls`` method for all nodes.

        The default implementation supports `.app` path. Your implementation
        should yield child nodes.

        Always call this as first before you yield your own methods to provide
        users with standard nodes.

        :param path: SHV path that should be listed.
        :return: Iterator over child nodes of the node on the path.
        """
        if not path:
            yield ".app"

    @staticmethod
    def _ls_node_for_path(
        path: str,
        full_path: str | collections.abc.Iterator[str],
    ) -> collections.abc.Iterator[str]:
        """Help with :meth:`_ls` implementation.

        This is helper designed to be called from :meth:`_ls` and simplifies a
        common operation that is extraction of the node name from the full path.

        With this function you effectively can remove the code like this from
        your codebase::

            fpath = "some/path/to/the/node"
            path = f"{path}/" if path else ""
            if fpath.startswith(path):
                yield fpath[len(path) :].partition("/")[0]

        And replace it with just::

            yield from self._ls_node_for_path(path, "some/path/to/the/node")

        :param path: ``path`` parameter passed to :meth:`_ls`.
        :param full_path: Path or iterator over paths the node name should be
          extracted from.
        """
        if path:
            path = f"{path}/"
        for pth in (full_path,) if isinstance(full_path, str) else full_path:
            if pth.startswith(path):
                yield pth[len(path) :].partition("/")[0]

    def _valid_path(self, path: str) -> bool:
        """Check that :meth:`_ls` reports this path as existing.

        :param path: SHV path to be validated.
        :return: ``True`` if :meth:`_ls` on parent node reports node with this
          tail piece. ``False`` is returned otherwise.
        """
        if not path:
            return True  # The root path always exists
        root, _, name = path.rpartition("/")
        return any(name == v for v in self._ls(root))

    def _method_call_dir(self, path: str, param: SHVType) -> SHVType:
        """Implement ``dir`` method call functionality."""
        if not self._valid_path(path):
            raise RpcMethodNotFoundError(f"No such node: {path}")
        if is_shvnull(param) or is_shvbool(param):
            return list(d.to_shv(bool(param)) for d in self._dir(path))
        if isinstance(param, str):
            return any(v.name == param for v in self._dir(path))
        raise RpcInvalidParamError("Use Null or Bool or String with node name")

    def _dir(self, path: str) -> collections.abc.Iterator[RpcDir]:  # noqa: PLR6301
        """Implement ``dir`` method for all nodes.

        This implementation is called only for valid paths
        (:meth:`shv.SHVBase._valid_path`).

        Always call this as first before you yield your own methods to provide
        users with standard ones.

        :param path: SHV path method should be listed for.
        :return: List of methods associated with given node.
        """
        yield RpcDir.stddir()
        yield RpcDir.stdls()
        if path == ".app":
            yield RpcDir.getter("shvVersionMajor", "n", "i")
            yield RpcDir.getter("shvVersionMinor", "n", "i")
            yield RpcDir.getter("name", "n", "s")
            yield RpcDir.getter("version", "n", "s")
            yield RpcDir.getter("date", "n", "t")
            yield RpcDir("ping")

    async def _got_signal(self, signal: Signal) -> None:
        """Handle signal.

        :param signal: SHV path to the node the signal is associated with.
        """

    class Request:
        """Set of parameters passed to the :meth:`shv.SHVBase._method_call`.

        This is provided as one data class to allow easier method typing as
        well as ability to more freely add or remove info items.
        """

        def __init__(self, msg: RpcMessage, shvbase: SHVBase) -> None:
            assert msg.type is RpcMessage.Type.REQUEST
            self._msg = msg
            self._shvbase = shvbase
            self._progress = 0.0
            self._progress_dirty = False

        @property
        def path(self) -> str:
            """SHV path to the node the method is associated with."""
            return self._msg.path

        @property
        def method(self) -> str:
            """SHV method name requested to be called."""
            return self._msg.method

        @property
        def param(self) -> SHVType:
            """Parameter provided for the method call."""
            return self._msg.param

        @property
        def access(self) -> RpcAccess:
            """Access level of the client specified in this request."""
            return self._msg.rpc_access or RpcAccess.BROWSE

        @property
        def user_id(self) -> str | None:
            """The user's ID collected as message was passed around.

            This can be ``None`` when request message contained no ID. You can raise
            :class:`RpcUserIDRequiredError` if you need it.
            """
            return self._msg.user_id

        @property
        def caller_ids(self) -> collections.abc.Sequence[int]:
            """The sequence of the IDs for the caller.

            This sequence uniquely identifies this caller for this
            application, thus it can be used to have unique context specific to
            the caller.
            """
            return self._msg.caller_ids

        @property
        def progress(self) -> float:
            """The progress signaled for this request."""
            return self._progress

        @progress.setter
        def progress(self, value: float) -> None:
            self._progress = value
            if not self._progress_dirty:
                self._progress_dirty = True
                self._shvbase._idle_message_ready()

        def ri_match(self, ri: str) -> bool:
            """Check if request matches given RI."""
            return rpcri_match(ri, self.path, self.method)

    class Signal:
        """Set of parameters passed to the :meth:`SHVBase._got_signal`.

        This is provided as one data class to allow easier method typing as
        well as ability to more freely add or remove info items.
        """

        def __init__(self, msg: RpcMessage) -> None:
            assert msg.type is RpcMessage.Type.SIGNAL
            self._msg = msg

        @property
        def path(self) -> str:
            """SHV path signal is associated with."""
            return self._msg.path

        @property
        def signal(self) -> str:
            """Signal name."""
            return self._msg.signal_name

        @property
        def source(self) -> str:
            """SHV method name signal is associated with."""
            return self._msg.source

        @property
        def param(self) -> SHVType:
            """Value caried by signal."""
            return self._msg.param

        @property
        def access(self) -> RpcAccess:
            """Access level of the signal."""
            return self._msg.rpc_access or RpcAccess.READ

        @property
        def user_id(self) -> str | None:
            """User's ID recorded in the signal.

            This can be ``None`` when request message contained no ID.
            """
            return self._msg.user_id

        def ri_match(self, ri: str) -> bool:
            """Check if signal matches given RI."""
            return rpcri_match(ri, self.path, self.source, self.signal)



"""Common tools for SHV RPC type definitions."""

import decimal


def strnum(num: int | None) -> str:
    """Convert number to string representation."""
    if num is None:
        return ""
    if abs(num) >= 100:  # Not worth it for two digits
        blen = num.bit_length()
        if blen == num.bit_count():
            return f">{'-' if num < 0 else ''}{blen}"
        if (1 << (blen - 1)) == abs(num):
            return f"^{'-' if num < 0 else ''}{blen - 1}"
    return str(num)


def strdec(num: decimal.Decimal | None) -> str:
    """Convert number to string representation."""
    if num is None:
        return ""
    result = str(num)
    if result.startswith("0."):
        return result[1:]
    if result.startswith("-0."):
        return f"-{result[2:]}"
    return result



"""The extension for the PurePosixPath to support our root."""

from __future__ import annotations

import collections.abc
import itertools
import os
import string
import typing

from .rpcri import shvpath_match


class SHVPath:
    """The SHV path helper.

    This is designed to be close to the :mod:`pathlib` while providing the SHV
    path specific behavior. But some of the methods just doesn't make sense in
    the respect of the SHV path (such as ``is_absolute``).

    We can't easily use :mod:`pathlib` based version because there are huge
    differences in the implementation (not external API) between Python 3.11
    and 3.12 and later versions.

    The rules for the SHV path are:

    - White characters (like spaces) are not allowed.
    - SHV path can't be absolute (can't start with ``/``).
    - ``""`` is parent of every path.
    - The combination of SHV path and string representing an absolute path
      (appending absolute path to SHV) results into only absolute path being
      used (while leading ``/`` is removed). This behavior is not derived from
      SHV path but rather from :mod:`pathlib`.
    - ``..`` is being interpreted as upped directory and thus removes upper
      node from the SHV path. This is extension of the standard SHV path
      behavior as SHV standard do not specify ``..`` interpretation.
    """

    def __init__(self, *pathsegments: str | os.PathLike) -> None:
        parts: list[str] = []
        for segment in itertools.chain.from_iterable(
            os.fspath(pth).split("/") for pth in pathsegments if pth
        ):
            if any(c in segment for c in string.whitespace):
                raise ValueError("SHV path can't contain white space characters")
            if segment == "..":
                if parts:
                    parts.pop()
            elif segment:
                parts.append(segment)
            else:
                parts = []
        self._parts = tuple(parts)

        self._parents: SHVPathParents | None = None
        self._hash: int | None = None

    @property
    def parts(self) -> tuple[str, ...]:
        """Tuple giving access to the path's various components."""
        return self._parts

    def __str__(self) -> str:
        return "/".join(self._parts)

    def __fspath__(self) -> str:
        return str(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self)!r})"

    def __reduce__(self) -> tuple[type[SHVPath], tuple[str, ...]]:
        return self.__class__, tuple(self._parts)  # pragma: no coverage

    def __bool__(self) -> bool:
        return bool(self._parts)

    def __truediv__(self, key: str | os.PathLike | SHVPath) -> SHVPath:
        return type(self)(*self._parts, key)

    def __rtruediv__(self, key: str | os.PathLike) -> SHVPath:
        return type(self)(key, *self._parts)

    def __hash__(self) -> int:
        if self._hash is None:
            self._hash = hash(tuple(self._parts))
        return self._hash

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str | os.PathLike):
            return os.fspath(self) == os.fspath(other)
        return super().__eq__(other)

    def __lt__(self, other: object) -> bool:
        if isinstance(other, str | os.PathLike):
            return os.fspath(self) < os.fspath(other)
        raise TypeError(f"unsupported '<' between '{type(self)}' and '{type(other)}'")

    def __le__(self, other: object) -> bool:
        if isinstance(other, str | os.PathLike):
            return os.fspath(self) <= os.fspath(other)
        raise TypeError(f"unsupported '<=' between '{type(self)}' and '{type(other)}'")

    def __gt__(self, other: object) -> bool:
        if isinstance(other, str | os.PathLike):
            return os.fspath(self) > os.fspath(other)
        raise TypeError(f"unsupported '>' between '{type(self)}' and '{type(other)}'")

    def __ge__(self, other: object) -> bool:
        if isinstance(other, str | os.PathLike):
            return os.fspath(self) >= os.fspath(other)
        raise TypeError(f"unsupported '>=' between '{type(self)}' and '{type(other)}'")

    @property
    def parents(self) -> SHVPathParents:
        """An immutable sequence providing access to the logical ancestors of the path."""
        if self._parents is None:
            self._parents = SHVPathParents(self)
        return self._parents

    @property
    def parent(self) -> SHVPath:
        """The logical parent of the path."""
        if self._parts:
            return type(self)(*self._parts[:-1])
        return self

    @property
    def name(self) -> str:
        """A string representing the final path component, if any."""
        return self._parts[-1] if self._parts else ""

    @property
    def suffix(self) -> str:
        """The last dot-separated portion of the final component, if any."""
        _, dot, suffix = self.name.rpartition(".")
        return f"{dot}{suffix}"

    @property
    def suffixes(self) -> list[str]:
        """The last dot-separated portion of the final component, if any."""
        return list(f".{s}" for s in self.name.split(".")[1:])

    @property
    def stem(self) -> str:
        """The final path component, without its suffix."""
        return self.name.partition(".")[0]

    def is_relative_to(self, path: str | os.PathLike | SHVPath) -> bool:
        """Return whether or not this path is relative to the other path."""
        if not isinstance(path, SHVPath):
            path = SHVPath(path)
        return path._parts == self._parts[: len(path._parts)]

    def joinpath(self, *pathsegments: str | os.PathLike) -> SHVPath:
        """Join paths together.

        Calling this method is equivalent to combining the path with each of
        the given path segments in turn.
        """
        return type(self)(*self._parts, *pathsegments)

    def full_match(self, pattern: str) -> bool:
        """Match this path against the provided glob-style pattern."""
        return shvpath_match(pattern, str(self))

    def match(self, pattern: str) -> bool:
        """Match this path against the provided glob-style pattern."""
        return any(
            shvpath_match(pattern, "/".join(parts))
            for parts in map(lambda i: self._parts[i:], range(len(self._parts)))
        )

    def relative_to(self, other: str | os.PathLike | SHVPath) -> SHVPath:
        """Compute a version of this path relative to the path represented by other."""
        if not isinstance(other, SHVPath):
            other = SHVPath(other)
        if self._parts[: len(other._parts)] != other._parts:
            raise ValueError(f"{self} is not relative to {other}")
        return SHVPath(*self._parts[len(other._parts) :])

    def with_name(self, name: str) -> SHVPath:
        """Return a new path with the name changed."""
        if not self._parts:
            raise ValueError("Root path has an empty name.")
        return type(self)(*self._parts[:-1], name)

    def with_stem(self, stem: str) -> SHVPath:
        """Return a new path with the stem changed."""
        if not self._parts:
            raise ValueError("Root path has an empty name.")
        _, dot, suffix = self._parts[-1].partition(".")
        return type(self)(*self._parts[:-1], f"{stem}{dot}{suffix}")

    def with_suffix(self, suffix: str) -> SHVPath:
        """Return a new path with the suffix changed."""
        if not self._parts:
            raise ValueError("Root path has an empty name.")
        return type(self)(*self._parts[:-1], f"{self.stem}{suffix}")


class SHVPathParents(collections.abc.Sequence[SHVPath]):
    """The parents generator for :class:`SHVPath`."""

    def __init__(self, path: SHVPath) -> None:
        self._path = path

    @typing.overload
    def __getitem__(self, i: int) -> SHVPath: ...

    @typing.overload
    def __getitem__(
        self, i: slice[typing.Any, typing.Any, typing.Any]
    ) -> collections.abc.Sequence[SHVPath]: ...

    def __getitem__(
        self, i: int | slice[typing.Any, typing.Any, typing.Any]
    ) -> collections.abc.Sequence[SHVPath] | SHVPath:
        parts = self._path.parts
        if isinstance(i, slice):
            return tuple(self[y] for y in range(*i.indices(len(parts))))

        if -len(parts) <= i < len(parts):
            return SHVPath(*parts[: -i - 1])
        raise IndexError("index out of range for SHV path parent")

    def __len__(self) -> int:
        return len(self._path.parts)



# SPDX-FileCopyrightText: 2025 Elektroline a.s.
#
# SPDX-License-Identifier: MIT

"""RPC Exchange usage and implementation utility."""

from __future__ import annotations

import abc
import collections.abc

from .rpcmethod import RpcMethodAccess, RpcMethodDesc
from .shvbase import SHVBase
from .value import SHVMapType, SHVType, is_shvmap


class RpcExchange:
    """RPC Exchange connection over :class:`SimpleBase`."""

    BLOCK_SIZE: int = 1024
    """The block size signaled as ready to be received.

    This can control how big and thus how many messages will be exchanged. It
    should be something reasonable, not that small (causing too many messages to
    be generated) as well as not too big (causing huge messages which is
    discouraged in SHV).
    """

    def __init__(
        self,
        client: SHVBase,
        path: str,
        idle_timeout: int = 30,
    ) -> None:
        self.client: SHVBase = client
        """The client used to access the RPC file."""
        self._path = path
        self._counter: int = 0
        self._to_receive: int = 0
        self._to_send: int = 0
        self._wbuf: bytearray = bytearray()
        self._rbuf: bytearray = bytearray()

    async def _exchange(self) -> None:
        """Perform exchange by sending wbuf and adding data to rbuf."""
        # TODO wait for signal in case we can't do anything instead of polling
        param: dict[int, SHVType] = {0: self._counter, 1: self.BLOCK_SIZE}
        if self._to_receive > 0:
            param[3] = self._wbuf[self._to_receive :]
            self._wbuf = self._wbuf[: self._to_receive]
        data = await self.client.call(self._path, "exchange", param)
        if not isinstance(data, bytes):
            raise Exception("Invalid response for exchange method received")
        self._rbuf += data
        self._counter += 1

    def _rbuf_take(self, n: int | None = None) -> bytes:
        result = self._rbuf[:n]
        del self._rbuf[:n]
        return result

    async def read(self, n: int = -1) -> bytes:
        """Read bytes from the exchange stream.

        This also sends any bytes that were previously written.

        :param n: Number of bytes to read. Zero or less is interpreted as
          unconstrained read (read everything until nothing is available to
          read at the moment).
        :return: Received bytes.
        """
        if n > 0:
            while len(self._rbuf) < n:
                await self._exchange()
        else:
            while True:
                prev_size = len(self._rbuf)
                await self._exchange()
                if prev_size == len(self._rbuf):
                    break
        return self._rbuf_take(n if n > 0 else None)

    async def readexactly(self, n: int) -> bytes:
        """Read exactly given number of bytes.

        See :meth:`read` for other implications on using this method.

        :param n: Number of bytes to be read.
        :return: Received bytes.
        """
        while len(self._rbuf) < n:
            await self._exchange()
        return self._rbuf_take(n)

    async def readuntil(self, separator: bytes = b"\n") -> bytes:
        """Read until given byte sequence is received.

        See :meth:`read` for other implications on using this method.

        :param separator: The separator to be located in the sequence.
        :return: Received bytes before separator and separator.
        """
        while separator not in self._rbuf:
            await self._exchange()
        return self._rbuf_take(self._rbuf.find(separator) + len(separator))

    def write(self, data: bytes) -> None:
        """Write bytes to the exchange stream.

        :param data: Data to be written to the exchange stream.
        """
        self._wbuf += data

    async def drain(self) -> None:
        """Drain all written bytes to the exchange stream.

        RPC Exchange is designed to do read and write with a single operation
        where reading is required to be performed before write is allowed. This
        means that this function must buffer read data. In most cases it is
        better to use :meth:`read`, :meth:`readexactly`, or :meth:`readuntil`
        directly instead of this one.
        """
        while self._wbuf:
            await self._exchange()

    async def options(self) -> SHVMapType:
        """Get the current options."""
        result = await self.client.call(self._path, "options")
        return result if is_shvmap(result) else {}

    async def set_options(self, options: SHVMapType) -> None:
        """Set the provided options."""
        await self.client.call(self._path, "setOptions", options)

    async def close(self) -> None:
        """Close the exchange stream."""
        await self.client.call(self._path, "close")

    @classmethod
    async def new(
        cls, client: SHVBase, path: str, options: SHVMapType | None = None
    ) -> RpcExchange:
        """Establish a new RPC Exchange connection."""
        token = await client.call(path, "newExchange", options)
        if not isinstance(token, str):
            raise Exception("Returned token of invalid type.")
        # if options and "idleTimeOut" in options: TODO
        return cls(client, f"{path}/{token}")


class RpcExchangeServer:
    """The server side for the RPC Exchange."""

    class Connection(abc.ABC):
        """The RPC exchange server connection."""

        def __init__(self, server: RpcExchangeServer) -> None:
            self.server = server
            """The reference to the server this connection belongs to."""

        @abc.abstractmethod
        async def exchange(self, data: bytes) -> bytes:
            """Perform the exchange operation."""

    def __init__(self) -> None:
        self._connections: dict[str, RpcExchangeServer.Connection] = {}

    def _ls(self, node: str) -> collections.abc.Iterator[str]:
        pass

    def dir(self, node: str) -> collections.abc.Iterator[RpcMethodDesc]:
        pass

    def method_call(
        self,
        name: str,
        param: SHVType,
        access_level: RpcMethodAccess,
    ) -> SHVType:
        pass



"""The SHV RPC type struct."""

from __future__ import annotations

import collections.abc
import typing

from .. import SHVIMapType, SHVType
from .base import RpcType


class RpcTypeStruct(RpcType, collections.abc.Mapping[int, tuple[RpcType, str]]):
    """The Struct type representation."""

    def __init__(
        self, items: collections.abc.Mapping[int, tuple[RpcType, str]]
    ) -> None:
        if len(items) != len(set(v[1] for v in items.values())):
            raise ValueError("Duplicate keys are not allowed")
        self._items = items
        self._index = {v[1]: k for k, v in items.items()}

    def key(self, key: str) -> int:
        """Convert string key to integer key."""
        return self._index[key]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RpcTypeStruct) and self._items == other._items

    def __str__(self) -> str:
        defs = (
            f"{tp}:{key}" if i is None else f"{tp}:{key}:{i}"
            for i, tp, key in self.__seqiter(iter(self._items.items()))
        )
        return f"i{{{','.join(defs)}}}"

    @staticmethod
    def __seqiter(
        iiter: collections.abc.Iterator[tuple[int, tuple[RpcType, str]]],
    ) -> collections.abc.Iterator[tuple[int | None, RpcType, str]]:
        expected = 0
        for k, v in iiter:
            yield None if expected == k else k, v[0], v[1]
            expected = k + 1

    def __getitem__(self, key: int) -> tuple[RpcType, str]:
        return self._items[key]

    def __iter__(self) -> collections.abc.Iterator[int]:
        return iter(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def validate(  # noqa: D102
        self, value: SHVType
    ) -> typing.TypeGuard[collections.abc.Mapping[int, SHVType]]:
        return (
            isinstance(value, collections.abc.Mapping)
            and all(k in self._items for k in value)
            and all(
                v[0].validate(typing.cast(SHVIMapType, value).get(k))
                for k, v in self._items.items()
            )
        )



"""Class that provides an easier way to define regular methods and signals."""

from __future__ import annotations

import asyncio
import collections
import collections.abc
import dataclasses
import typing
import weakref

from .rpcaccess import RpcAccess
from .rpcdir import RpcDir
from .rpcerrors import RpcInvalidParamError, RpcUserIDRequiredError
from .rpcmessage import RpcMessage
from .shvbase import SHVBase
from .value import SHVType

SHVMethodT: typing.TypeAlias = collections.abc.Callable[
    [typing.Any, SHVBase.Request],
    SHVType | collections.abc.Coroutine[None, None, SHVType],
]
SHVGetMethodT: typing.TypeAlias = collections.abc.Callable[
    [typing.Any, int | None],
    collections.abc.Coroutine[None, None, SHVType] | SHVType,
]
SHVSetMethodT: typing.TypeAlias = collections.abc.Callable[
    [typing.Any, SHVType, str | None],
    collections.abc.Coroutine[None, None, None] | None,
]


class SHVMethods(SHVBase):
    """SHV RPC methods and signals implementation helper.

    The regular way to implement methods in :class:`shv.SHVBase` is by defining
    :meth:`shv.SHVBase._ls`, :meth:`shv.SHVBase._dir`, and
    :meth:`shv.SHVBase._method_call` methods. That provides ability to implement
    discoverability as well as method implementation in various dynamic ways.
    But in real world applications it is common that we just want to define
    custom methods and having to implement these methods can be error-prone and
    unnecessary complex. This class instead provides a way to define methods
    that are discovered and their integration is handled automatically.
    """

    def __init__(
        self,
        *args: typing.Any,  # noqa ANN401
        **kwargs: typing.Any,  # noqa ANN401
    ) -> None:
        self._methods: dict[str, dict[str, SHVMethods.Method]] = (
            collections.defaultdict(dict)
        )
        for name in dir(self):
            if isinstance(attr := getattr(self, name, None), self.Method):
                if attr.desc.name in self._methods[attr.path]:
                    raise ValueError(
                        f"Method already defined {attr.path}:{attr.desc.name}"
                    )
                self._methods[attr.path][attr.desc.name] = attr
        super().__init__(*args, **kwargs)

    async def _method_call(self, request: SHVBase.Request) -> SHVType:
        if (
            request.path in self._methods
            and request.method in self._methods[request.path]
        ):
            method = self._methods[request.path][request.method]
            if request.access >= method.desc.access:
                if (
                    RpcDir.Flag.USER_ID_REQUIRED in method.desc.flags
                    and request.user_id is None
                ):
                    raise RpcUserIDRequiredError
                return await method._func_call(request)
        return await super()._method_call(request)

    def _ls(self, path: str) -> collections.abc.Iterator[str]:
        """Implement `ls` method for registered methods."""
        yield from super()._ls(path)
        yield from self._ls_node_for_path(path, iter(self._methods))

    def _dir(self, path: str) -> collections.abc.Iterator[RpcDir]:
        """Provide method description for registered methods."""
        yield from super()._dir(path)
        if path in self._methods:
            for method in self._methods[path].values():
                yield method.desc

    @dataclasses.dataclass(frozen=True)
    class Method:
        """The definition of the RPC Method.

        This allows you to define methods as attributes of the object and such
        methods are automatically discovered and used. Note that the discovery
        is performed on object initialization.

        The following is provided and thus not required to be performed in the
        method itself:

        - Check if SHV path and method name matches as otherwise this method
          would not be called.
        - Verification of the access level.
        - Verifies if user's ID is required (based on the
          :attr:`shv.RpcDir.flags`) and sends
          :class:`shv.RpcUserIDRequiredError` error.

        What application has to do on its own:

        - Implement the method behavior and return the method result.
        - Validate parameter and raise :class:`shv.RpcInvalidParamError` for
          invalid values.
        - Explicitly send signals.
        """

        path: str
        """The SHV path this method is associated with."""
        desc: RpcDir
        """Method description.

        This defines method name, acceess level as well as other options that
        are considered on method call request.
        """
        func: SHVMethodT
        """Python method providing the implementation."""

        _bound: weakref.ReferenceType[SHVMethods] | None = None

        def __get__(self, instance: SHVMethods, owner: object) -> SHVMethods.Method:
            """Object method bounding."""
            return dataclasses.replace(self, _bound=weakref.ref(instance))

        async def signal(
            self,
            value: SHVType = None,
            signal: str | None = None,
            access: RpcAccess = RpcAccess.READ,
            user_id: str | None = None,
        ) -> None:
            """External call is used to send signals."""
            shvmethods = self._bound() if self._bound is not None else None
            if shvmethods is None:
                raise UnboundLocalError
            if signal is None:
                if not self.desc.signals:
                    raise NotImplementedError("Method doesn't have any signals")
                signal = next(iter(self.desc.signals.keys()))
            elif signal not in self.desc.signals:
                raise ValueError(f"Invalid signal name: {signal}")
            await shvmethods._send(
                RpcMessage.signal(
                    self.path, signal, self.desc.name, value, access, user_id
                )
            )

        async def _func_call(self, request: SHVBase.Request) -> SHVType:
            shvmethods = self._bound() if self._bound is not None else None
            if shvmethods is None:
                raise UnboundLocalError
            res = self.func(shvmethods, request)
            if asyncio.iscoroutine(res):
                res = await res
            return res

    @classmethod
    def method(
        cls, path: str, desc: RpcDir
    ) -> collections.abc.Callable[[SHVMethodT], SHVMethods.Method]:
        """Decorate method to turn it to :class:`shv.SHVMethods.Method`.

        :param path: SHV path this method is associated with.
        :param desc: Method description.
        """
        return lambda func: cls.Method(path, desc, func)

    @classmethod
    def property(
        cls,
        path: str,
        tp: str = "Any",
        access: RpcAccess = RpcAccess.READ,
        signal: bool | str = False,
    ) -> collections.abc.Callable[[SHVGetMethodT], SHVMethods.Method]:
        """Decorate method to use method as ``get`` property method.

        The properties are common standard ways to organize access to the values
        exposed over SHV. This decorator provides an easier way to implement
        these common methods by validating the parameter.

        The decorated function is called in wrapper that on top of the
        :class:`Method` handling also verifies the parameter.

        :param path: SHV path for this property.
        :param tp: Type description provided by this property.
        :param access: Minimal access level required for `get` method.
        :param signal: If property has signal associated with it and optionally
          its name. `chng` is used if `True` is provided and `False` (the]
          default) is used for no signal at all.
        """

        def wrapper(func: SHVGetMethodT) -> SHVMethods.Method:
            def substitut(
                obj: SHVMethods, request: SHVBase.Request
            ) -> collections.abc.Coroutine[None, None, SHVType] | SHVType:
                if request.param is not None and not isinstance(request.param, int):
                    raise RpcInvalidParamError("Only Int or Null allowed")
                return func(obj, request.param)

            return cls.Method(
                path,
                RpcDir.getter(result=tp, access=access, signal=signal),
                substitut,
            )

        return wrapper

    @classmethod
    def property_setter(
        cls,
        getter: Method,
        access: RpcAccess = RpcAccess.WRITE,
    ) -> collections.abc.Callable[[SHVSetMethodT], SHVMethods.Method]:
        """Decorate method to make it set method for the property node.

        This is a companion decorator to the :meth:`property` that allows
        addition of the ``set`` method to the property node.

        :param getter: The reference to the ``get`` method.
        :param access: Minimal access level required for ``set`` method.
        """
        if getter.desc.name != "get":
            raise ValueError("Can be used only on method get")

        def wrapper(func: SHVSetMethodT) -> SHVMethods.Method:
            def substitut(
                obj: SHVMethods, request: SHVBase.Request
            ) -> collections.abc.Coroutine[None, None, None] | None:
                return func(obj, request.param, request.user_id)

            return cls.Method(
                getter.path,
                RpcDir.setter(param=getter.desc.result, access=access),
                substitut,
            )

        return wrapper



"""SHV RPC type Enum."""

from __future__ import annotations

import collections.abc
import typing

from .. import SHVType
from .base import RpcType


class RpcTypeEnum(RpcType, collections.abc.Mapping[int, str]):
    """The Unsigned type representation."""

    def __init__(self, items: collections.abc.Mapping[int, str]) -> None:
        self._items = items
        if len(items) != len(set(items.values())):
            raise ValueError("Duplicates are not allowed in enum")

    def __str__(self) -> str:
        defs = (
            self._items[i] if seq else f"{self._items[i]}:{i}"
            for seq, i in self.__seqiter(iter(self._items))
        )
        return f"i[{','.join(defs)}]"

    @staticmethod
    def __seqiter(
        iiter: collections.abc.Iterator[int],
    ) -> collections.abc.Iterator[tuple[bool, int]]:
        expected = 0
        for i in iiter:
            yield expected == i, i
            expected = i + 1

    def __getitem__(self, key: int) -> str:
        return self._items[key]

    def __iter__(self) -> collections.abc.Iterator[int]:
        return iter(self._items)

    def __len__(self) -> int:
        return len(self._items)

    def validate(self, value: SHVType) -> typing.TypeGuard[int]:  # noqa: D102
        return isinstance(value, int) and value in self
